from typing import Any, List, Union, Dict
from typing_extensions import TypedDict, Literal

DataSource = Literal["issues", "boards"]

JiraFieldName = str


class SelectFieldDefinition(TypedDict):
    expression: str
    column: str


Field = Union[SelectFieldDefinition, JiraFieldName]

JQLString = str

CustomFilterFieldName = str

JqlList = List[str]

Expression = str

ExpressionList = List[Expression]

# We have to use the alternative method of defining the TypedDict since
# one of our dictionary fields' names is a reserved word
QueryDefinition = TypedDict(
    "QueryDefinition",
    {
        "select": List[Field],
        "from": DataSource,
        "where": JqlList,
        "order_by": JqlList,
        "having": ExpressionList,
        "group_by": ExpressionList,
        "sort_by": ExpressionList,
        "expand": List[str],
        "limit": int,
        "cap": int,
        "cache": int,
    },
    total=False,
)


class SchemaRow(TypedDict, total=False):
    id: str
    type: str
    description: str
    raw: Any


class ViewerDefinitionDict(TypedDict, total=False):
    csv: str


class ShellConfigDict(TypedDict, total=False):
    emacs_mode: bool


class InstanceDefinition(TypedDict, total=False):
    url: str
    username: str
    password: str


class ConfigDict(TypedDict, total=False):
    instances: Dict[str, InstanceDefinition]
    shell: ShellConfigDict
    viewers: ViewerDefinitionDict



import argparse
import subprocess
import sys
import tempfile
from typing import IO, Dict, Optional, cast

from yaml import safe_load

from ..exceptions import UserError
from ..plugin import BaseCommand, get_installed_formatters
from ..query import Executor
from ..types import QueryDefinition


class Command(BaseCommand):
    DEFAULT_VIEWERS: Dict[str, str] = {"csv": "vd"}

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        formatters = get_installed_formatters()

        parser.add_argument("query_file", help="Query definition file to run")
        parser.add_argument(
            "--format",
            "-f",
            choices=formatters.keys(),
            default="csv",
            help="Data format in which to write record data; default: 'csv'.",
        )
        parser.add_argument(
            "--output",
            "-o",
            help="Path to file where records will be written; default: stdout.",
        )
        parser.add_argument(
            "--view",
            "-v",
            default=False,
            action="store_true",
            help="Launch viewer immediately after completing query.",
        )

    @classmethod
    def get_help(cls) -> str:
        return "Interactively generates a query definition (in yaml format)."

    def handle(self) -> None:
        viewer: Optional[str] = cast(
            str, self.config.get("viewers", {}).get(self.options.format)
        ) or self.DEFAULT_VIEWERS.get(self.options.format)
        if not viewer and self.options.view:
            raise UserError(f"No viewer set for format {self.options.format}")
        formatter_cls = get_installed_formatters()[self.options.format]

        query_definition: QueryDefinition = {}
        with open(self.options.query_file, "r") as inf:
            query_definition = safe_load(inf)

        output: IO[str] = sys.stdout
        output_file = self.options.output
        if self.options.output:
            output = open(self.options.output, "w")
        elif self.options.view:
            output = tempfile.NamedTemporaryFile(
                "w", suffix=f".{formatter_cls.get_file_extension()}"
            )
            output_file = output.name

        query = Executor(
            self.jira, query_definition, progress_bar=output is not sys.stdout
        )
        with formatter_cls(query, output) as formatter:
            for row in query:
                formatter.writerow(row)
        output.flush()

        if self.options.view:
            assert viewer

            proc = subprocess.Popen([viewer, output_file])
            proc.wait()

        if output is not sys.stdout:
            output.close()



from typing import Any, List

from ..plugin import BaseFunction


class Function(BaseFunction):
    def __call__(self, scalar: List[Any], dotpath: str) -> Any:  # type: ignore[override]
        values = []

        path_parts = dotpath.split(".")

        for item in scalar:
            if item is None:
                continue

            try:
                cursor = item
                for part in path_parts:
                    if isinstance(cursor, dict):
                        cursor = cursor[part]
                    else:
                        cursor = getattr(cursor, part)
            except (KeyError, AttributeError):
                continue

            if cursor is not None:
                values.append(cursor)

        return values



from __future__ import annotations

from abc import ABCMeta, abstractmethod
import argparse
import copy
import json
import logging
import random
import statistics
from typing import IO, TYPE_CHECKING, Any, Callable, Dict, Optional, Type, cast

import pkg_resources

from dateutil.parser import parse as parse_datetime
import keyring
from jira import JIRA
from rich.console import Console

from .constants import APP_NAME
from .exceptions import ConfigurationError
from .types import ConfigDict, InstanceDefinition
from .utils import save_config, get_functions_for_module

if TYPE_CHECKING:
    from .query import Executor

logger = logging.getLogger(__name__)


BUILTIN_FUNCTIONS: Dict[str, Callable] = {
    # Built-ins
    "abs": abs,
    "all": all,
    "any": any,
    "bin": bin,
    "bool": bool,
    "hex": hex,
    "int": int,
    "len": len,
    "max": max,
    "min": min,
    "oct": oct,
    "ord": ord,
    "pow": pow,
    "range": range,
    "reversed": reversed,
    "round": round,
    "set": set,
    "sorted": sorted,
    "str": str,
    "sum": sum,
    "tuple": tuple,
    "map": map,
    "filter": filter,
    # Statistics
    **get_functions_for_module(
        statistics,
        [
            "fmean",
            "geometric_mean",
            "harmonic_mean",
            "mean",
            "median",
            "median_grouped",
            "median_high",
            "median_low",
            "mode",
            "multimode",
            "pstdev",
            "pvariance",
            "quantiles",
            "stdev",
            "variance",
        ],
    ),
    # Random
    **get_functions_for_module(random, ["random", "randrange", "randint", "choice"],),
    # JSON
    "json_loads": json.loads,
    "json_dumps": json.dumps,
    # Date Handling
    "parse_datetime": parse_datetime,
}
REGISTERED_FUNCTIONS: Dict[str, Callable] = {}


def get_installed_commands() -> Dict[str, Type[BaseCommand]]:
    possible_commands: Dict[str, Type[BaseCommand]] = {}
    for entry_point in pkg_resources.iter_entry_points(group="jira_select.commands"):
        try:
            loaded_class = entry_point.load()
        except ImportError:
            logger.warning(
                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
                entry_point,
            )
            continue
        if not issubclass(loaded_class, BaseCommand):
            logger.warning(
                "Loaded entrypoint %s, but loaded class is "
                "not a subclass of `jira_select.plugin.BaseCommand`.",
                entry_point,
            )
            continue
        possible_commands[entry_point.name] = loaded_class

    return possible_commands


class BaseCommand(metaclass=ABCMeta):
    _jira: Optional[JIRA] = None

    def __init__(self, config: ConfigDict, options: argparse.Namespace):
        self._config: ConfigDict = config
        self._options: argparse.Namespace = options
        self._console = Console(highlight=False)
        super().__init__()

    @property
    def config(self) -> ConfigDict:
        """ Provides the configuration dictionary."""
        return self._config

    def save_config(self) -> None:
        """ Saves the existing configuration dictionary."""
        save_config(self.config, self.options.config)

    @property
    def options(self) -> argparse.Namespace:
        """ Provides options provided at the command-line."""
        return self._options

    @property
    def console(self) -> Console:
        """ Provides access to the console (see `rich.console.Console`."""
        return self._console

    @property
    def jira(self) -> JIRA:
        """ Provides access to the configured Jira instance."""
        if self._jira is None:
            instance: Dict[InstanceDefinition] = cast(  # type: ignore
                InstanceDefinition,
                self.config.get("instances", {}).get(self.options.instance_name, {}),
            )

            instance_url = self.options.instance_url or instance.get("url")
            if not instance_url:
                raise ConfigurationError(
                    "instance_url not set; please run `jira-select configure`."
                )

            username = self.options.username or instance.get("username")
            if not username:
                raise ConfigurationError(
                    "username not set; please run `jira-select configure`."
                )

            password = self.options.password or instance.get("password")
            if not password:
                password = keyring.get_password(APP_NAME, instance_url + username)
                if not password:
                    raise ConfigurationError(
                        f"Password not stored for {instance_url} user {username}; "
                        "use the 'store-password' command to store the password "
                        "for this user account in your system keyring or use "
                        "`jira-select configure`."
                    )
            self._jira = JIRA(
                server=instance_url,
                options={"agile_rest_path": "agile",},
                basic_auth=(username, password),
            )

        return self._jira

    @classmethod
    def get_help(cls) -> str:
        """ Retuurns help text for this function."""
        return ""

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        """ Allows adding additional command-line arguments. """
        pass

    @abstractmethod
    def handle(self) -> None:
        """ This is where the work of your function starts. """
        ...


def get_installed_formatters() -> Dict[str, Type[BaseFormatter]]:
    possible_formatters: Dict[str, Type[BaseFormatter]] = {}
    for entry_point in pkg_resources.iter_entry_points(group="jira_select.formatters"):
        try:
            loaded_class = entry_point.load()
        except ImportError:
            logger.warning(
                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
                entry_point,
            )
            continue
        if not issubclass(loaded_class, BaseFormatter):
            logger.warning(
                "Loaded entrypoint %s, but loaded class is "
                "not a subclass of `jira_select.plugin.BaseFormatter`.",
                entry_point,
            )
            continue
        possible_formatters[entry_point.name] = loaded_class

    return possible_formatters


class BaseFormatter(metaclass=ABCMeta):
    def __init__(self, executor: Executor, stream: IO[str]):
        self._executor = executor
        self._stream = stream

    @classmethod
    @abstractmethod
    def get_file_extension(cls) -> str:
        ...

    @property
    def executor(self):
        return self._executor

    @property
    def stream(self):
        return self._stream

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        # Exception handling here
        self.close()

    def open(self):
        return

    def close(self):
        return

    @abstractmethod
    def writerow(self, row: Dict[str, Any]):
        ...


def register_function(name: str, fn: Callable):
    """ Register a callable to be a function available in queries."""
    REGISTERED_FUNCTIONS[name] = fn


def get_installed_functions(jira: JIRA = None) -> Dict[str, Callable]:
    possible_commands: Dict[str, Callable] = copy.copy(BUILTIN_FUNCTIONS)
    possible_commands.update(REGISTERED_FUNCTIONS)

    for entry_point in pkg_resources.iter_entry_points(group="jira_select.functions"):
        try:
            loaded_class = entry_point.load()
        except ImportError:
            logger.warning(
                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
                entry_point,
            )
            continue
        if not issubclass(loaded_class, BaseFunction):
            logger.warning(
                "Loaded entrypoint %s, but loaded class is "
                "not a subclass of `jira_select.plugin.BaseFunction`.",
                entry_point,
            )
            continue
        possible_commands[entry_point.name] = cast(Callable, loaded_class(jira))

    return possible_commands


class BaseFunction(metaclass=ABCMeta):
    def __init__(self, jira: Optional[JIRA]):
        self._jira = jira

    @property
    def jira(self):
        assert self._jira

        return self._jira

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Optional[Any]:
        ...



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from functools import total_ordering
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    cast,
    Iterable,
    Tuple,
    Type,
    Union,
)

from jira import JIRA, Issue
from rich.progress import Progress, TaskID
from diskcache import Cache


from .plugin import get_installed_functions
from .types import (
    ExpressionList,
    JqlList,
    QueryDefinition,
    SelectFieldDefinition,
    Expression,
)
from .utils import (
    calculate_result_hash,
    expression_includes_group_by,
    get_cache_path,
    get_field_data,
    get_row_dict,
    parse_sort_by_definition,
    parse_select_definition,
)


@total_ordering
class NullAcceptableSort:
    def __init__(self, value: Any):
        self._value = value

    def __le__(self, other):
        if other._value is None:
            return True
        if self._value is None:
            return False

        return self._value < other._value

    def __eq__(self, other):
        if self._value is other._value:
            return True

        try:
            return self._value == other._value
        except Exception:
            return False

    def __str__(self):
        return str(self._value)


class Result(metaclass=ABCMeta):
    def __getattr__(self, name):
        result = self.as_dict()

        if name not in result:
            raise AttributeError(name)

        return result[name]

    @abstractmethod
    def as_dict(self) -> Dict[str, Any]:
        ...

    @abstractmethod
    def single(self) -> SingleResult:
        ...

    def evaluate_expression(
        self,
        expression: Expression,
        group_by: ExpressionList = None,
        functions: Dict[str, Callable] = None,
    ):
        if group_by is None:
            group_by = []

        row_source = self

        # Force returning a single value if the expression under
        # evaluation is one we're grouping on
        if expression_includes_group_by(expression, group_by):
            row_source = self.single()

        return get_field_data(row_source, expression, functions)


class SingleResult(Result):
    def __init__(self, row: Any):
        self._row = row

    def as_dict(self) -> Dict[str, Any]:
        return get_row_dict(self._row)

    def single(self) -> SingleResult:
        return self


class GroupedResult(Result):
    def __init__(
        self, rows: List[SingleResult] = None,
    ):
        super().__init__()
        self._rows: List[SingleResult] = rows if rows is not None else []

    @property
    def rows(self):
        return self._rows

    @property
    def all_fields(self):
        fields = set()

        for row in self._rows:
            fields |= set(row.as_dict())

        return fields

    def add(self, record: SingleResult):
        self._rows.append(record)

    def single(self) -> SingleResult:
        return self.rows[0]

    def as_dict(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {}

        for field in self.all_fields:
            # Exclude empty rows -- in SQL when you run an aggregation
            # function on a set of rows, NULL rows are skipped, so we
            # should probably do the same?
            result[field] = [
                row.as_dict()[field]
                for row in self._rows
                if row.as_dict().get(field) is not None
            ]

        return result


class Query:
    def __init__(self, jira: JIRA, definition: QueryDefinition):
        self._jira = jira
        self._definition = definition

    def _ensure_str(self, iterable=Iterable[Any]) -> List[str]:
        return [str(item) for item in iterable]

    def _get_all_fields(self):
        fields = [{"id": "key"}, {"id": "id"}] + self._jira.fields()
        for field in fields:
            definition: SelectFieldDefinition = {
                "expression": field["id"],
                "column": field["id"],
            }
            yield definition

    @property
    def select(self) -> List[SelectFieldDefinition]:
        fields: List[SelectFieldDefinition] = []

        for field in self._definition["select"]:
            if field == "*":
                fields.extend(self._get_all_fields())
            else:
                fields.append(parse_select_definition(field))

        return fields

    @property
    def from_(self) -> str:
        return self._definition["from"]

    @property
    def where(self) -> JqlList:
        return self._ensure_str(self._definition.get("where", []))

    @property
    def order_by(self) -> JqlList:
        return self._ensure_str(self._definition.get("order_by", []))

    @property
    def having(self) -> ExpressionList:
        return self._ensure_str(self._definition.get("having", []))

    @property
    def group_by(self) -> ExpressionList:
        return self._ensure_str(self._definition.get("group_by", []))

    @property
    def sort_by(self) -> List[Tuple[Expression, bool]]:
        return [
            parse_sort_by_definition(definition)
            for definition in self._ensure_str(self._definition.get("sort_by", []))
        ]

    @property
    def expand(self) -> List[str]:
        return self._ensure_str(self._definition.get("expand", []))

    @property
    def limit(self) -> Optional[int]:
        return self._definition.get("limit")

    @property
    def cap(self) -> Optional[int]:
        return self._definition.get("cap")

    @property
    def cache(self) -> Optional[int]:
        return self._definition.get("cache")


class NullProgressbar:
    def update(self, *args, **kwargs):
        pass

    def add_task(self, *args, **kwargs) -> TaskID:
        return TaskID(0)

    def remove_task(self, *args, **kwargs):
        pass

    def __enter__(self) -> NullProgressbar:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        pass


class NullCache:
    def __init__(self, *args, **kwargs):
        pass

    def __getitem__(self, key):
        raise KeyError(key)

    def get(self, key, default=None, **kwargs):
        return default

    def touch(self, key, **kwargs):
        return

    def set(self, key, value, **kwargs):
        return

    def add(self, key, value, **kwargs):
        return


class Executor:
    def __init__(
        self,
        jira: JIRA,
        definition: QueryDefinition,
        enable_cache: bool = True,
        progress_bar: bool = False,
    ):
        self._query: Query = Query(jira, definition)
        # self._definition: QueryDefinition = clean_query_definition(definition)
        self._jira: JIRA = jira
        self._functions: Dict[str, Callable] = get_installed_functions(jira)
        self._progress_bar_enabled = progress_bar

        # Set a high ceiling at the start so the progressbar does not
        # start "full"
        self._jira_count = 2 ** 32
        self._group_by_count = 2 ** 32
        self._having_count = 2 ** 32
        self._sort_by_count = 2 ** 32

        self._cache: Union[Cache, NullCache] = NullCache()
        if enable_cache:
            self._cache = Cache(get_cache_path())

    @property
    def jira(self) -> JIRA:
        return self._jira

    @property
    def cache(self) -> Union[Cache, NullCache]:
        return self._cache

    @property
    def query(self) -> Query:
        return self._query

    @property
    def functions(self) -> Dict[str, Callable]:
        return self._functions

    def _get_jql(self) -> str:
        query = " AND ".join(self.query.where)
        order_by_fields = ", ".join(self.query.order_by)

        if order_by_fields:
            query = f"{query} ORDER BY {order_by_fields}"

        return query

    def _get_issues(self, task: TaskID) -> Generator[SingleResult, None, None]:
        jql = self._get_jql()
        cache_key = f"{jql}:{','.join(self.query.order_by)}:{self.query.limit}"

        start_at = 0
        max_results = 2 ** 32
        result_limit = self.query.limit or 2 ** 32

        if self.query.cache:
            try:
                cached_results = self.cache[cache_key]
                self._jira_count = len(cached_results)
                self.progress.remove_task(task)
                for result in cached_results:
                    yield SingleResult(Issue({}, None, result))
                return
            except KeyError:
                pass

        cache = []

        while start_at < min(max_results, result_limit):
            results = self.jira.search_issues(
                jql,
                startAt=start_at,
                expand=",".join(self.query.expand),
                fields="*all",
                maxResults=min(result_limit, 100),
            )
            max_results = results.total
            self._jira_count = min(results.total, result_limit)
            for result in results:
                result = cast(Issue, result)
                self.progress.update(task, advance=1, total=self.jira_count)

                if self.query.cache:
                    cache.append(result.raw)

                start_at += 1
                yield SingleResult(result)

                # Return early if our result limit has been reached
                if start_at >= result_limit:
                    break

        if self.query.cache:
            self.cache.set(cache_key, cache, expire=self.query.cache)

    @property
    def jira_count(self):
        return self._jira_count

    def _process_group_by(
        self, iterator: Generator[SingleResult, None, None], task: Optional[TaskID]
    ) -> Generator[Result, None, None]:
        groups: Dict[int, GroupedResult] = {}
        self._group_by_count = 0

        if not self.query.group_by:
            for row in iterator:
                self._group_by_count = self.jira_count
                yield row
            return

        assert task is not None

        for row in iterator:
            self.progress.update(task, total=self.jira_count)

            row_hash = calculate_result_hash(row, self.query.group_by, self.functions,)
            if row_hash not in groups:
                self._group_by_count += 1
                groups[row_hash] = GroupedResult()

            groups[row_hash].add(row)

            self.progress.update(task, advance=1)

        for _, value in groups.items():
            yield value

    @property
    def group_by_count(self):
        return self._group_by_count

    def _process_having(
        self, iterator: Generator[Result, None, None], task: Optional[TaskID]
    ) -> Generator[Result, None, None]:
        self._having_count = 0
        if not self.query.having:
            for row in iterator:
                self._having_count = self.group_by_count
                yield row
            return

        assert task is not None

        for row in iterator:
            self.progress.update(task, total=self.group_by_count)

            include_row = True
            for having in self.query.having:
                if not row.evaluate_expression(
                    having, self.query.group_by, self.functions,
                ):
                    include_row = False
                    break

            if include_row:
                self._having_count += 1
                yield row

            self.progress.update(task, advance=1)

    @property
    def having_count(self):
        return self._having_count

    def _process_sort_by(
        self, iterator: Generator[Result, None, None], task: Optional[TaskID],
    ) -> Generator[Result, None, None]:
        self._sort_by_count = 0
        if not self.query.sort_by:
            for row in iterator:
                self._sort_by_count = self.having_count
                yield row
            return

        assert task is not None

        # First, materialize our list
        rows = list(iterator)
        self.progress.update(task, total=len(rows))
        self._sort_by_count = len(rows)

        # Now, sort by each of the ordering expressions in reverse order
        for sort_expression, reverse in reversed(self.query.sort_by):

            def sort_key(row):
                result = row.evaluate_expression(
                    sort_expression, self.query.group_by, self.functions
                )
                self.progress.update(task, advance=1)

                return NullAcceptableSort(result)

            rows = sorted(rows, key=sort_key, reverse=reverse)

        yield from rows

    @property
    def sort_by_count(self):
        return self._sort_by_count

    def _generate_row_dict(self, row: Result) -> Dict[str, Any]:
        result: Dict[str, Any] = {}

        for field_defn in self.query.select:
            result[field_defn["column"]] = row.evaluate_expression(
                field_defn["expression"], self.query.group_by, self.functions,
            )

        return result

    def _get_iterator(self) -> Generator[Dict[str, Any], None, None]:
        iterator_task = self.progress.add_task("jira", total=2 ** 32)
        group_by_task: Optional[TaskID] = None
        if self.query.group_by:
            group_by_task = self.progress.add_task("group_by", total=2 ** 32)
        having_task: Optional[TaskID] = None
        if self.query.having:
            having_task = self.progress.add_task("having", total=2 ** 32)
        sort_by_task: Optional[TaskID] = None
        if self.query.sort_by:
            sort_by_task = self.progress.add_task("sort_by", total=2 ** 32)
        select_task = self.progress.add_task("select", total=2 ** 32)

        if self.query.from_ == "issues":
            iterator = self._get_issues
        else:
            raise NotImplementedError(
                f"No search for source {self.query.from_} implemented."
            )

        for row in self._process_sort_by(
            self._process_having(
                self._process_group_by(iterator(iterator_task), group_by_task),
                having_task,
            ),
            sort_by_task,
        ):
            self.progress.update(select_task, total=self.sort_by_count)
            yield self._generate_row_dict(row)
            self.progress.update(select_task, advance=1)

    @property
    def progress(self) -> Union[Progress, NullProgressbar]:
        return self._progress_bar

    def __iter__(self) -> Generator[Dict[str, Any], None, None]:
        progress_bar_cls: Union[Type[Progress], Type[NullProgressbar]] = NullProgressbar
        if self._progress_bar_enabled:
            progress_bar_cls = Progress

        with progress_bar_cls() as progress:
            self._progress_bar = progress

            row_count = 0
            for row in self._get_iterator():
                yield row

                row_count += 1
                if self.query.cap is not None and row_count >= self.query.cap:
                    break



from __future__ import annotations

import hashlib
import logging
import os
import re
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union, Tuple

from appdirs import user_config_dir
from jira.resources import Resource
import simpleeval
from simpleeval import EvalWithCompoundTypes
from yaml import safe_dump, safe_load

from .constants import APP_NAME
from .exceptions import QueryError
from .types import (
    ConfigDict,
    ExpressionList,
    SelectFieldDefinition,
    Expression,
)

if TYPE_CHECKING:
    from .query import Result

FIELD_DISPLAY_DEFN_RE = re.compile(r'^(?P<expression>.*) as "(?P<column>.*)"$')
SORT_BY_DESC_FN = re.compile(r"^(?P<expression>.*) DESC", re.IGNORECASE)
SORT_BY_ASC_FN = re.compile(r"^(?P<expression>.*) ASC", re.IGNORECASE)


logger = logging.getLogger(__name__)


def get_config_dir() -> str:
    root_path = user_config_dir(APP_NAME, "coddingtonbear")
    os.makedirs(root_path, exist_ok=True)

    return root_path


def get_cache_path() -> str:
    config_path = get_config_dir()
    cache_path = os.path.join(config_path, "cache")
    os.makedirs(cache_path, exist_ok=True)

    return cache_path


def get_default_config_path() -> str:
    root_path = get_config_dir()
    return os.path.join(root_path, "config.yaml",)


def get_config(path: str = None) -> ConfigDict:
    if path is None:
        path = get_default_config_path()

    if not os.path.isfile(path):
        return {}

    with open(path, "r") as inf:
        return safe_load(inf)


def save_config(data: ConfigDict, path: str = None) -> None:
    if path is None:
        path = get_default_config_path()

    with open(path, "w") as outf:
        safe_dump(data, outf)


def get_functions_for_module(
    module: ModuleType, names: List[str]
) -> Dict[str, Callable]:
    """ Return functions in module matching the specified name.

    This exists because some functions we want to publish as available
    functions are available only on Python >= 3.8.

    """
    functions: Dict[str, Callable] = {}

    for name in names:
        if hasattr(module, name):
            functions[name] = getattr(module, name)

    return functions


def parse_select_definition(
    expression: Union[str, SelectFieldDefinition]
) -> SelectFieldDefinition:
    if isinstance(expression, str):
        as_match = FIELD_DISPLAY_DEFN_RE.match(expression)
        column = expression
        expression = expression
        if as_match:
            match_dict = as_match.groupdict()
            expression = match_dict["expression"]
            column = match_dict["column"]

        return {"expression": expression, "column": column}
    return expression


def parse_sort_by_definition(expression: str) -> Tuple[str, bool]:
    is_reversed = False
    is_reversed_match = SORT_BY_DESC_FN.match(expression)
    is_not_reversed_match = SORT_BY_ASC_FN.match(expression)
    if is_reversed_match:
        expression = is_reversed_match.groupdict()["expression"]
        is_reversed = True
    elif is_not_reversed_match:
        expression = is_not_reversed_match.groupdict()["expression"]
        is_reversed = False

    return expression, is_reversed


def calculate_result_hash(
    row: Result, group_fields: ExpressionList, functions: Dict[str, Callable],
) -> int:
    params = [
        str(get_field_data(row, group_field, functions)) for group_field in group_fields
    ]

    return int(hashlib.sha1(":".join(params).encode("UTF-8")).hexdigest(), 16)


def get_row_dict(row: Any) -> Dict[str, Any]:
    names: Dict[str, Any] = {}

    if hasattr(row, "fields"):
        for field_name in dir(row.fields):
            if not field_name.startswith("_"):
                names[field_name] = getattr(row.fields, field_name)

    # Gather any top-level keys, too, to make sure we fetch any expansions
    for key in dir(row):
        value = getattr(row, key)
        if (
            key != "fields"
            and not key.startswith("_")
            and not key.upper() == key
            and not callable(value)
        ):
            names[key] = getattr(row, key)

    return names


def expression_includes_group_by(
    expression: Expression, group_by: List[Expression]
) -> bool:
    for group_by_expression in group_by:
        if group_by_expression in expression:
            return True

    return False


def evaluate_expression(
    expression: str,
    names: Dict[str, Any],
    functions: Optional[Dict[str, Callable]] = None,
) -> Any:
    return EvalWithCompoundTypes(names=names, functions=functions).eval(expression)


def normalize_value(value: Any) -> Any:
    if isinstance(value, (str, bool, float, int)):
        return value
    elif isinstance(value, list):
        return [normalize_value(v) for v in value]
    elif isinstance(value, Dict):
        return {k: normalize_value(v) for k, v in value.items()}
    elif isinstance(value, Resource):
        display = str(value)
        if display.startswith("<JIRA"):
            # If the display value starts with "<JIRA", we know
            # the library couldn't find a pretty way of printing
            # this field; let's just return the raw dictionary
            # in that case
            return value.raw
        return display
    elif "PropertyHolder" in str(value):
        # Yes; this LOOKS ridiculous, but in older versions
        # of the Jira library, the PropertyHolder class is
        # created a runtime :facepalm:.  I think this is
        # the only way to find them :shrug:
        return {
            name: normalize_value(getattr(value, name))
            for name in dir(value)
            if not name.startswith("_") and not name.upper() == name
        }

    return value


def get_field_data(
    row: Result,
    expression: str,
    functions: Optional[Dict[str, Callable]] = None,
    error_returns_null=True,
) -> Any:
    if functions is None:
        functions = {}

    try:
        return normalize_value(
            evaluate_expression(
                expression, names={"_": row, **row.as_dict()}, functions=functions,
            )
        )
    except (
        simpleeval.AttributeDoesNotExist,
        simpleeval.NameNotDefined,
        KeyError,
        IndexError,
        TypeError,
        AttributeError,
    ):
        if not error_returns_null:
            raise
        return None
    except Exception as e:
        raise QueryError(str(e))



import re
from typing import Optional

from ..plugin import BaseFunction


class Function(BaseFunction):
    """ Converts a string estimate to days."""

    def __call__(self, value_str: str, day_hour_count=8) -> Optional[float]:  # type: ignore[override]
        unit_mapping = {
            "w": 5,
            "d": 1,
            "h": 1 / day_hour_count,
            "m": 1 / day_hour_count * 60,
        }
        unit_matcher = re.compile(r"(?P<value>[\d.]+)(?P<unit>[^\d]+)")
        total_value: float = 0

        for value_str_unit in value_str.split(" "):
            match = unit_matcher.match(value_str_unit)
            if not match:
                raise ValueError(value_str)

            unit = match.groupdict()["unit"]
            value = float(match.groupdict()["value"])

            try:
                total_value += unit_mapping[unit] * value
            except KeyError:
                pass

        return total_value



import argparse
from typing import Any, Callable, Dict, Iterable

from dotmap import DotMap
from rich.table import Table
from simpleeval import NameNotDefined

from ..exceptions import JiraSelectError
from ..plugin import BaseCommand, get_installed_functions
from ..types import SchemaRow
from ..utils import evaluate_expression


class Command(BaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.functions: Dict[str, Callable] = get_installed_functions(self.jira)

    @classmethod
    def get_help(cls) -> str:
        return "Searches Jira for fields matching your search query."

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "source", choices=["issues"],
        )
        parser.add_argument(
            "search_terms",
            nargs="*",
            type=str,
            help="Case-insensitive search term for limiting displayed results.",
        )
        parser.add_argument(
            "--having",
            help=(
                "A 'having' expression to use for limiting the displayed results. "
                "Searchable fields are 'key', 'type', and 'description' and 'raw'. "
                "E.g.--having=\"'estimate' in description.lower()\". "
            ),
        )

    def evaluate_expression(self, row: Any, expression: str) -> Any:
        return evaluate_expression(expression, row, functions=self.functions)

    def get_field_data(self, row: Any, expression: str) -> str:
        result = self.evaluate_expression(row, expression)
        if isinstance(result, str):
            return result
        return ""

    def get_issue_data(self) -> Iterable[SchemaRow]:
        for column in sorted(self.jira.fields(), key=lambda field: field["name"]):
            try:
                type = str(self.get_field_data(DotMap(column), "schema.type"))
            except NameNotDefined:
                type = ""
            data: SchemaRow = {
                "id": str(self.get_field_data(DotMap(column), "id")),
                "type": type,
                "description": str(self.get_field_data(DotMap(column), "name")),
                "raw": DotMap(column),
            }
            yield data

    def handle(self) -> None:
        source_data_fns: Dict[str, Callable[[], Iterable[SchemaRow]]] = {
            "issues": self.get_issue_data
        }

        try:
            data_fn = source_data_fns[self.options.source]
        except KeyError:
            raise JiraSelectError(
                f"No source data function defined for {self.options.source}"
            )

        table = Table(title=self.options.source)
        table.add_column(header="id", style="green")
        table.add_column(header="type", style="cyan")
        table.add_column(header="description", style="bright_cyan")

        for row in data_fn():
            if self.options.search_terms:
                matches = True
                for option in self.options.search_terms:
                    if option.lower() not in str(row).lower():
                        matches = False
                        break
                if not matches:
                    continue
            if self.options.having:
                if not self.evaluate_expression(DotMap(row), self.options.having):
                    continue

            table.add_row(
                row["id"], row["type"], row["description"],
            )

        self.console.print(table)
